import ij.IJ;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.GenericDialog;
import ij.gui.Roi;
import ij.measure.Calibration;
import ij.plugin.PlugIn;
import ij.process.ImageProcessor;
import imagescience.images.ColorImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.Image;
import imagescience.transforms.Crop;
import imagescience.utilities.I5DResource;
import imagescience.utilities.Messenger;
import java.awt.Rectangle;
import java.awt.image.ColorModel;

public class TJ_Crop implements PlugIn {

    private static String xrange = "0,0";
    private static String yrange = "0,0";
    private static String zrange = "1,1";
    private static String trange = "1,1";
    private static String crange = "1,1";
    
    public void run(String arg) {

	msg.debug(TJ_Options.debug);
	msg.debug("TransformJ "+TJ_Options.VERSION+": Crop");

	if (IJ.versionLessThan("1.35i")) return;

	final ImagePlus imp = WindowManager.getCurrentImage();
	if (imp == null)  { error("There are no images open"); return; }

	final Roi roi = imp.getRoi();
	if (roi != null) {
	    final Rectangle rect = roi.getBounds();
	    xrange = rect.x + "," + (rect.x + rect.width - 1);
	    yrange = rect.y + "," + (rect.y + rect.height - 1);
	    zrange = "1," + imp.getNSlices();
	    trange = "1," + imp.getNFrames();
	    crange = "1," + imp.getNChannels();
	}

	GenericDialog gd = new GenericDialog("TransformJ: Crop");
	gd.addStringField("x-range for cropping:",xrange,10);
	gd.addStringField("y-range for cropping:",yrange,10);
	gd.addStringField("z-range for cropping:",zrange,10);
	gd.addStringField("t-range for cropping:",trange,10);
	gd.addStringField("ch-range for cropping:",crange,10);
	gd.showDialog();

	if (gd.wasCanceled()) return;

	xrange = gd.getNextString();
	yrange = gd.getNextString();
	zrange = gd.getNextString();
	trange = gd.getNextString();
	crange = gd.getNextString();

	try {
	    int startx, starty, startz, startt, startc;
	    int stopx, stopy, stopz, stopt, stopc;
	    try {
		startx = Integer.parseInt(xrange.substring(0,xrange.indexOf(',')));
		stopx = Integer.parseInt(xrange.substring(xrange.indexOf(',')+1,xrange.length()));
	    } catch (Exception e) { throw new IllegalArgumentException("Invalid x-range for cropping"); }
	    try {
		starty = Integer.parseInt(yrange.substring(0,yrange.indexOf(',')));
		stopy = Integer.parseInt(yrange.substring(yrange.indexOf(',')+1,yrange.length()));
	    } catch (Exception e) { throw new IllegalArgumentException("Invalid y-range for cropping"); }
	    try {
		startz = Integer.parseInt(zrange.substring(0,zrange.indexOf(','))) - 1;
		stopz = Integer.parseInt(zrange.substring(zrange.indexOf(',')+1,zrange.length())) - 1;
	    } catch (Exception e) { throw new IllegalArgumentException("Invalid z-range for cropping"); }
	    try {
		startt = Integer.parseInt(trange.substring(0,trange.indexOf(','))) - 1;
		stopt = Integer.parseInt(trange.substring(trange.indexOf(',')+1,trange.length())) - 1;
	    } catch (Exception e) { throw new IllegalArgumentException("Invalid t-range for cropping"); }
	    try {
		startc = Integer.parseInt(crange.substring(0,crange.indexOf(','))) - 1;
		stopc = Integer.parseInt(crange.substring(crange.indexOf(',')+1,crange.length())) - 1;
	    } catch (Exception e) { throw new IllegalArgumentException("Invalid ch-range for cropping"); }

	    final Image img = Image.wrap(imp);
	    final Crop cropper = new Crop();
	    cropper.debug(TJ_Options.debug);
	    final Coordinates startpos = new Coordinates(startx,starty,startz,startt,startc);
	    final Coordinates stoppos = new Coordinates(stopx,stopy,stopz,stopt,stopc);
	    final Image newimg = cropper.run(img,startpos,stoppos);
	    ImagePlus newimp = newimg.imageplus();
	    final Calibration cal = imp.getCalibration().copy();
	    newimp.setCalibration(cal);
	    final ImageProcessor ip = imp.getProcessor();
	    final ImageProcessor newip = newimp.getProcessor();
	    final ColorModel cm = ip.getColorModel();
	    newip.setColorModel(cm);
	    newimp.getStack().setColorModel(cm);
	    newip.setMinAndMax(ip.getMin(),ip.getMax());
	    boolean i5dinstalled = false;
	    try { Class.forName("Image5D"); i5dinstalled = true; } catch (Exception e) { }
	    if (i5dinstalled && !(newimg instanceof ColorImage)) {
		final Dimensions newdims = newimg.dimensions();
		final Dimensions indims = img.dimensions();
		if (newdims.t > 1 || newdims.s > 1) {
		    newimp = I5DResource.convert(newimp,true);
		    if (I5DResource.instance(imp)) {
			final int len = stopc - startc + 1;
			final int[][] idx = new int[2][len];
			for (int i=0; i<len; ++i) {
			    idx[0][i] = startc + i + 1;
			    idx[1][i] = i + 1;
			}
			I5DResource.transfer(imp,newimp,idx);
		    }
		} else if (indims.t > 1 || indims.s > 1) {
		    if (I5DResource.instance(imp))
			I5DResource.transfer(imp,newimp,startc+1);
		}
	    }
	    newimp.changes = TJ_Options.asksave;
	    newimp.show();

	} catch (OutOfMemoryError e) {
	    error("Not enough memory for this operation");

	} catch (IllegalArgumentException e) {
	    error(e.getMessage());
	}
    }

    private void error(final String message) { msg.error("TransformJ: Error", message+"."); }

    private final Messenger msg = new Messenger();

}
